@@ -9,21 +9,6 @@ from eventkit import Event
 from .contract import Contract, ScanData, TagValue
 from .util import EPOCH, UNSET_DOUBLE, UNSET_INTEGER
 
-__all__ = (
-    'SoftDollarTier PriceIncrement Execution CommissionReport '
-    'BarDataList RealTimeBarList BarData RealTimeBar '
-    'HistogramData NewsProvider DepthMktDataDescription '
-    'ScannerSubscription ScanDataList FundamentalRatios '
-    'ExecutionFilter PnL PnLSingle AccountValue TickData '
-    'TickByTickAllLast TickByTickBidAsk TickByTickMidPoint '
-    'HistoricalTick HistoricalTickBidAsk HistoricalTickLast '
-    'TickAttrib TickAttribBidAsk TickAttribLast '
-    'MktDepthData DOMLevel TradeLogEntry '
-    'FamilyCode SmartComponent '
-    'PortfolioItem Position Fill OptionComputation OptionChain Dividends '
-    'NewsArticle HistoricalNews NewsTick NewsBulletin ConnectionStats'
-).split()
-
 nan = float('nan')
 
 
@@ -199,6 +184,33 @@ class PnLSingle:
     value: float = nan
 
 
+@dataclass
+class HistoricalSession:
+    startDateTime: str = ''
+    endDateTime: str = ''
+    refDate: str = ''
+
+
+@dataclass
+class HistoricalSchedule:
+    startDateTime: str = ''
+    endDateTime: str = ''
+    timeZone: str = ''
+    sessions: List[HistoricalSession] = field(default_factory=list)
+
+
+@dataclass
+class WshEventData:
+    conId: int = UNSET_INTEGER
+    filter: str = ''
+    fillWatchlist: bool = False
+    fillPortfolio: bool = False
+    fillCompetitors: bool = False
+    startDate: str = ''
+    endDate: str = ''
+    totalLimit: int = UNSET_INTEGER
+
+
 class AccountValue(NamedTuple):
     account: str
     tag: str
