@@ -12,23 +12,20 @@ from ib_insync.contract import (
     Contract, ContractDescription, ContractDetails, DeltaNeutralContract,
     ScanData)
 from ib_insync.objects import (
-    AccountValue, BarData, BarDataList, CommissionReport,
-    DOMLevel, DepthMktDataDescription,
-    Dividends, Execution, Fill, FundamentalRatios, HistogramData,
-    HistoricalNews, HistoricalTick, HistoricalTickBidAsk, HistoricalTickLast,
-    MktDepthData, NewsArticle, NewsBulletin, NewsProvider, NewsTick,
-    OptionChain, OptionComputation, PnL, PnLSingle, PortfolioItem, Position,
-    PriceIncrement, RealTimeBar, RealTimeBarList, TickAttribBidAsk,
-    TickAttribLast, TickByTickAllLast, TickByTickBidAsk, TickByTickMidPoint,
-    TickData, TradeLogEntry)
+    AccountValue, BarData, BarDataList, CommissionReport, DOMLevel,
+    DepthMktDataDescription, Dividends, Execution, FamilyCode, Fill,
+    FundamentalRatios, HistogramData, HistoricalNews, HistoricalSchedule,
+    HistoricalSession, HistoricalTick, HistoricalTickBidAsk,
+    HistoricalTickLast, MktDepthData, NewsArticle, NewsBulletin, NewsProvider,
+    NewsTick, OptionChain, OptionComputation, PnL, PnLSingle, PortfolioItem,
+    Position, PriceIncrement, RealTimeBar, RealTimeBarList, SoftDollarTier,
+    TickAttribBidAsk, TickAttribLast, TickByTickAllLast, TickByTickBidAsk,
+    TickByTickMidPoint, TickData, TradeLogEntry)
 from ib_insync.order import Order, OrderState, OrderStatus, Trade
 from ib_insync.ticker import Ticker
 from ib_insync.util import (
     UNSET_DOUBLE, UNSET_INTEGER, dataclassAsDict, dataclassUpdate,
-    globalErrorEvent, isNan, parseIBDatetime)
-
-__all__ = ['RequestError', 'Wrapper']
-
+    getLoop, globalErrorEvent, isNan, parseIBDatetime)
 
 OrderKeyType = Union[int, Tuple[int, int]]
 
@@ -87,8 +84,6 @@ class Wrapper:
         self.ticker2ReqId: Dict[Union[int, str], Dict[Ticker, int]] = \
             defaultdict(dict)
         #    tickType -> Ticker -> reqId
-        self.reqId2MarketDataType: Dict[int, int] = {}
-        #   reqId -> marketDataType
         self.reqId2Subscriber: Dict[int, Any] = {}
         #    live bars or live scan data
         self.reqId2PnL: Dict[int, PnL] = {}
@@ -113,11 +108,19 @@ class Wrapper:
         self._timeout: float = 0
         self.setTimeout(0)
 
+    def setEventsDone(self):
+        """Set all subscription-type events as done."""
+        events = [ticker.updateEvent for ticker in self.tickers.values()]
+        events += [sub.updateEvent for sub in self.reqId2Subscriber.values()]
+        for trade in self.trades.values():
+            events += [
+                trade.statusEvent, trade.modifyEvent, trade.fillEvent,
+                trade.filledEvent, trade.commissionReportEvent,
+                trade.cancelEvent, trade.cancelledEvent]
+        for event in events:
+            event.set_done()
+
     def connectionClosed(self):
-        for ticker in self.tickers.values():
-            ticker.updateEvent.set_done()
-        for sub in self.reqId2Subscriber.values():
-            sub.updateEvent.set_done()
         error = ConnectionError('Socket disconnect')
         for future in self._futures.values():
             if not future.done():
@@ -128,7 +131,7 @@ class Wrapper:
     def startReq(self, key, contract=None, container=None):
         """
         Start a new request and return the future that is associated
-        with with the key and container. The container is a list by default.
+        with the key and container. The container is a list by default.
         """
         future = asyncio.Future()
         self._futures[key] = future
@@ -212,7 +215,7 @@ class Wrapper:
         if not delay:
             delay = self._timeout - diff
         if delay > 0:
-            loop = asyncio.get_event_loop()
+            loop = getLoop()
             self._timeoutHandle = loop.call_later(delay, self._setTimer)
         else:
             self._logger.debug('Timeout')
@@ -301,6 +304,14 @@ class Wrapper:
     def positionEnd(self):
         self._endReq('positions')
 
+    def positionMulti(
+            self, reqId: int, account: str, modelCode: str,
+            contract: Contract, pos: float, avgCost: float):
+        pass
+
+    def positionMultiEnd(self, reqId: int):
+        pass
+
     def pnl(
             self, reqId: int, dailyPnL: float, unrealizedPnL: float,
             realizedPnL: float):
@@ -339,7 +350,7 @@ class Wrapper:
         """
         if order.whatIf:
             # response to whatIfOrder
-            if orderState.commissionCurrency:
+            if orderState.initMarginChange != str(UNSET_DOUBLE):
                 self._endReq(order.orderId, orderState)
         else:
             key = self.orderKey(order.clientId, order.orderId, order.permId)
@@ -350,6 +361,7 @@ class Wrapper:
                 trade.order.lmtPrice = order.lmtPrice
                 trade.order.auxPrice = order.auxPrice
                 trade.order.orderType = order.orderType
+                trade.order.orderRef = order.orderRef
             else:
                 # ignore '?' values in the order
                 order = Order(**{
@@ -369,7 +381,7 @@ class Wrapper:
                 # response to reqOpenOrders or reqAllOpenOrders
                 results.append(order)
 
-        # make sure that the client issues order ids larger then any
+        # make sure that the client issues order ids larger than any
         # order id encountered (even from other clients) to avoid
         # "Duplicate order id" error
         self.ib.client.updateReqId(orderId + 1)
@@ -519,7 +531,9 @@ class Wrapper:
         self._endReq(f'marketRule-{marketRuleId}', priceIncrements)
 
     def marketDataType(self, reqId: int, marketDataId: int):
-        self.reqId2MarketDataType[reqId] = marketDataId
+        ticker = self.reqId2Ticker.get(reqId)
+        if ticker:
+            ticker.marketDataType = marketDataId
 
     def realtimeBar(
             self, reqId: int, time: int, open_: float, high: float, low: float,
@@ -656,7 +670,6 @@ class Wrapper:
         if price or size:
             tick = TickData(self.lastTime, tickType, price, size)
             ticker.ticks.append(tick)
-        ticker.marketDataType = self.reqId2MarketDataType.get(reqId, 0)
         self.pendingTickers.add(ticker)
 
     def tickSize(self, reqId: int, tickType: int, size: float):
@@ -701,6 +714,8 @@ class Wrapper:
             ticker.auctionVolume = size
         elif tickType == 36:
             ticker.auctionImbalance = size
+        elif tickType == 61:
+            ticker.regulatoryImbalance = size
         elif tickType == 86:
             ticker.futuresOpenInterest = size
         elif tickType == 87:
@@ -710,7 +725,6 @@ class Wrapper:
         if price or size:
             tick = TickData(self.lastTime, tickType, price, size)
             ticker.ticks.append(tick)
-        ticker.marketDataType = self.reqId2MarketDataType.get(reqId, 0)
         self.pendingTickers.add(ticker)
 
     def tickSnapshotEnd(self, reqId: int):
@@ -776,7 +790,13 @@ class Wrapper:
         if not ticker:
             return
         try:
-            if tickType == 47:
+            if tickType == 32:
+                ticker.bidExchange = value
+            elif tickType == 33:
+                ticker.askExchange = value
+            elif tickType == 84:
+                ticker.lastExchange = value
+            elif tickType == 47:
                 # https://interactivebrokers.github.io/tws-api/fundamental_ratios_tags.html
                 d = dict(t.split('=')                     # type: ignore
                          for t in value.split(';') if t)  # type: ignore
@@ -784,8 +804,8 @@ class Wrapper:
                     with suppress(ValueError):
                         if v == '-99999.99':
                             v = 'nan'
-                        d[k] = float(v)
-                        d[k] = int(v)
+                        d[k] = float(v)                   # type: ignore
+                        d[k] = int(v)                     # type: ignore
                 ticker.fundamentalRatios = FundamentalRatios(**d)
             elif tickType in (48, 77):
                 # RT Volume or RT Trade Volume string format:
@@ -864,7 +884,15 @@ class Wrapper:
     def tickReqParams(
             self, reqId: int, minTick: float, bboExchange: str,
             snapshotPermissions: int):
-        pass
+        ticker = self.reqId2Ticker.get(reqId)
+        if not ticker:
+            return
+        ticker.minTick = minTick
+        ticker.bboExchange = bboExchange
+        ticker.snapshotPermissions = snapshotPermissions
+
+    def smartComponents(self, reqId, components):
+        self._endReq(reqId, components)
 
     def mktDepthExchanges(
             self, depthMktDataDescriptions: List[DepthMktDataDescription]):
@@ -1026,9 +1054,27 @@ class Wrapper:
             dividendsToLastTradeDate: float):
         pass
 
-    def error(self, reqId: int, errorCode: int, errorString: str):
+    def historicalSchedule(
+            self, reqId: int, startDateTime: str, endDateTime: str,
+            timeZone: str, sessions: List[HistoricalSession]):
+        schedule = HistoricalSchedule(
+            startDateTime, endDateTime, timeZone, sessions)
+        self._endReq(reqId, schedule)
+
+    def userInfo(self, reqId: int, whiteBrandingId: str):
+        self._endReq(reqId)
+
+    def softDollarTiers(self, reqId: int, tiers: List[SoftDollarTier]):
+        pass
+
+    def familyCodes(self, familyCodes: List[FamilyCode]):
+        pass
+
+    def error(
+            self, reqId: int, errorCode: int, errorString: str,
+            advancedOrderRejectJson: str):
         # https://interactivebrokers.github.io/tws-api/message_codes.html
-        warningCodes = {165, 202, 399, 404, 434, 492, 10167}
+        warningCodes = {110, 165, 202, 399, 404, 434, 492, 10167}
         isWarning = errorCode in warningCodes or 2100 <= errorCode < 2200
         msg = (
             f'{"Warning" if isWarning else "Error"} '
@@ -1052,6 +1098,8 @@ class Wrapper:
             elif (self.clientId, reqId) in self.trades:
                 # something is wrong with the order, cancel it
                 trade = self.trades[(self.clientId, reqId)]
+                if advancedOrderRejectJson:
+                    trade.advancedError = advancedOrderRejectJson
                 if not trade.isDone():
                     status = trade.orderStatus.status = OrderStatus.Cancelled
                     logEntry = TradeLogEntry(
